from lextok.rules._pattern import (
    CM,
    CODE,
    CONNECTOR,
    COURT,
    ENTITY_RULER_LABELS,
    OF,
    OF_THE_PH_,
    PROV_DIGITS,
    SPAN_RULER_LABELS,
    TH,
    THE,
    VS,
    DigitLists,
    Label,
    Rule,
    _orth_in,
    _re,
    camel_case_pattern,
    lower_words,
    name_code,
    name_court,
    name_statute,
    titled_words,
    uncamel,
)

from .abbreviations import Abbv, Prov
from .attribute_ruler import Attr, WordAttributes
from .citeable_builder import Duo, Style
from .custom_tokenizer import (
    INFIXES_OVERRIDE,
    create_special_rules,
    custom_prefix_list,
    custom_suffix_list,
)
from .entity_rules_citeable import (
    CourtName,
    DocketNum,
    ProvisionNum,
    ReporterNum,
    StatuteNum,
)



from collections.abc import Iterator
from enum import Enum
from typing import Any

from lextok.rules._pattern import (
    CM,
    CONNECTOR,
    COURT,
    DigitLists,
    Label,
    Rule,
    _multi,
    _orth_in,
    _re,
    name_court,
    titled_words,
)
from lextok.rules.abbreviations import Prov
from lextok.rules.citeable_builder import Style


class ReporterNum(Enum):
    scra = ["SCRA"]
    phil = ["Phil.", "Phil"]
    og = ["OG", "O.G.", "Off. Gaz."]


reporter_nums = Rule(
    label=Label.ReporterNum,
    patterns=[
        [{"LIKE_NUM": True}, {"ORTH": style}, {"LIKE_NUM": True}]
        for member in ReporterNum
        for style in member.value
    ],
)


class CourtName(Enum):
    SC = [{"LOWER": "supreme", "IS_TITLE": True}, COURT]
    RTC = titled_words("regional trial court")
    CA = name_court("appeals")
    CFI = name_court("first instance")
    MeTC = titled_words("metropolitan trial court")
    MCTC = titled_words("municipal circuit court")
    MTC = titled_words("municipal trial court")
    MTCC = (
        titled_words("municipal trial court")
        + [{"LOWER": "in"}]
        + titled_words("the cities")
    )


court_names = Rule(label=Label.CourtName, patterns=[mem.value for mem in CourtName])  # type: ignore


class DocketNum(Enum):
    GR = Style(let="gr")
    AM = Style(
        let="am",
        v=[
            "adm mat",
            "adm. mat",
            "adm mat.",
            "adm. mat.",
            "adm matter",
            "adm. matter",
            "admin. matter",
            "admin matter",
            "administrative matter",
        ],
    )
    AC = Style(
        let="ac",
        v=[
            "adm case",
            "adm. case",
            "admin. case",
            "admin case",
            "administrative case",
        ],
    )
    BM = Style(let="bm", v=["bar mat.", "bar matter"])


ENT_DOCKET_NUM = Rule(
    label=Label.DocketNum, patterns=[p for mem in DocketNum for p in mem.value.patterns]
)


class StatuteNum(Enum):
    RA = Style(let="ra", r="\\d{1,5}", v=["republic act", "rep. act"])
    CA = Style(let="ca", r="\\d{1,3}", v=["commonwealth act", "com. act"])
    BP = Style(let="bp", r="\\d{1,3}(-?(A|B))?", v=["batas pambansa"], pre=["Blg."])
    EO = Style(let="eo", r="\\d{1,4}(-?(A|B|C))?", v=["executive order", "exec. order"])
    PD = Style(
        let="pd",
        r="\\d{1,4}(-?(A|B))?",
        v=["presidential decree", "pres. decree", "pres. dec."],
    )
    ACT = Style(r="\\d{1,4}", v=["act"])


statute_nums = Rule(
    label=Label.StatuteNum,
    patterns=[p for mem in StatuteNum for p in mem.value.patterns],
)


class ProvisionNum(Enum):
    """A statute's text is divided into (sometimes hierarchical) provisions.
    This structure combines both the adjective division, e.g. `Sec.`, `Section`, `Bk.`, etc.
    (which may have different casing and abbreviations) with presumed valid serial numbers,
    e.g. `1`, `1-b`, `III`, etc.
    """

    abbreviated_Sec1 = list(Prov.set_abbvs())
    abbreviated_sec1 = list(Prov.set_abbvs(cased="lower"))
    abbreviated_SEC1 = list(Prov.set_abbvs(cased="upper"))
    Section1 = list(Prov.set_fulls())
    section1 = list(Prov.set_fulls(cased="lower"))
    SECTION1 = list(Prov.set_fulls(cased="upper"))

    @classmethod
    def generate(cls) -> Iterator[list[dict[str, Any]]]:
        specific_digits = _re("(" + "|".join(DigitLists.generate_options()) + ")")
        for member in cls:
            for end in [{"POS": "NUM"}, specific_digits]:
                start = [CONNECTOR]  # start
                if member.name.startswith("abbreviated_"):
                    yield start + [_orth_in(member.value), {"ORTH": "."}, end]
                    yield start + [_orth_in([f"{v}." for v in member.value]), end]
                yield start + [_orth_in(member.value), end]


special_clauses: list[dict[str, Any]] = [
    Label.ORDINAL.opt,
    {"LOWER": {"IN": ["whereas"]}},
    {"LOWER": {"IN": ["clause", "clauses"]}},
]
provision_nums = Rule(
    label=Label.ProvisionNum, patterns=list(ProvisionNum.generate()) + [special_clauses]
)
num_sym = _orth_in(["No", "Nos", "No.", "Nos."])
opt_nums = {"POS": {"IN": ["NUM", "PUNCT"]}, "OP": "*"}

ENT_GENERIC_NUM = Rule(
    label=Label.GenericNum,
    patterns=[[num_sym, {"POS": "NUM"}]],
)
"""Create a generic serial number pattern which hasn't been previously defined as an entity."""


provs = {"ENT_TYPE": Label.ProvisionNum.name, "OP": "+"}
stat = {
    "ENT_TYPE": {
        "IN": [
            Label.StatuteNamed.name,
            Label.StatuteNum.name,
            Label.DocketNum.name,
        ]
    }
}
opt_comma = CM | {"OP": "?"}
opt_comma_plus = {"ORTH": {"IN": ["of", "the", ","]}, "OP": "*"}

linker = {"ORTH": {"IN": ["or", "and", ",", "&"]}, "OP": "+"}

multi_linked_statute_pattern = _multi([linker, stat], 10)  # type: ignore
for linked_list in multi_linked_statute_pattern:
    linked_list.insert(0, stat)
as_amended: list[dict[str, Any]] = [
    {"ORTH": ",", "IS_PUNCT": True, "OP": "?"},
    {"LOWER": "as", "OP": "?"},
    {"LOWER": "amended"},
    {"LOWER": "by", "OP": "?"},
]

linked_statutes = Rule(
    label=Label.StatutoryLink,
    patterns=multi_linked_statute_pattern + [[stat] + as_amended + [stat]],
)
"""Connect statutory numbers together: `RA 141, RA 4124, RA 5325`"""

statutory_provisions = Rule(
    label=Label.StatutoryProvision,
    patterns=[
        [provs, opt_comma_plus, stat],
        [provs, opt_comma_plus, stat],
        [stat, opt_comma_plus, provs],
    ],
)
"""Connect statutes with provisions: `Art. 2, Civil Code`"""

decision_citations = Rule(
    label=Label.DecisionCitation,
    patterns=[
        [
            Label.CaseName.opt,
            Label.DocketNum.node,
            opt_nums,
            Label.DATE.node,
            opt_comma,
            Label.ReporterNum.node,
        ],
        [
            Label.CaseName.opt,
            Label.DocketNum.node,
            opt_nums,
            Label.DATE.node,
        ],
        [
            Label.CaseName.node,
            Label.ReporterNum.node,
            opt_nums,
            opt_comma,
            Label.DATE.node,
        ],
    ],
)
"""Connect decision names, docket citations, date, and/or reporter numbers: `X v. Y, GR No. 3425, Jan. 1, 2000, 14 SCRA 14`"""



import importlib
import io
import itertools
import re
import string
from collections.abc import Iterator
from enum import Enum, auto
from pathlib import Path
from typing import Any

import inflect
import jsonlines
import roman  # type: ignore
from pydantic import BaseModel, model_serializer
from slugify import slugify
from spacy.tokens import Doc, Span
from spacy.util import filter_spans

camel_case_pattern = re.compile(
    r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)"
)


def uncamel(text: str) -> list[str]:
    """For text in camelCaseFormatting, convert into a list of strings."""
    return [m.group(0) for m in camel_case_pattern.finditer(text)]


class Label(Enum):
    """A `Label` is shorthand for creating a spacy `ENT_TYPE` for entity-based patterns
    and `LABEL`s for span patterns. An alternate route would be to use 2 types of
    data structures (one for entities and one for spans) but this adds complexity downstream
    since all `Rule`s will now have to have different label fields rather than just a single one.

    Each member of the `Label` Enum will contain properties and methods:

    1. `@node` to represent `{"ENT_TYPE": <Enum member name>}`
    2. `@opt`: adds `{"OP": "?}` to the `@node`
    3. `@snakecase`: each label gets converted to a snakecased variable that can be used as a custom Doc attribute
    4. `extract_entities(doc)`: get all Spans that are entities a _member_ `Label`
    5. `extract_spans(doc)`: : get all Spans that match a SpanRuler key which is a _member_ `Label`
    """

    ORDINAL = auto()
    PERSON = auto()
    GPE = auto()
    LAW = auto()
    ORG = auto()
    DATE = auto()
    Actor = auto()
    GovtDivision = auto()
    CourtName = auto()
    Concept = auto()  # span
    Doctrine = auto()  # span
    Paper = auto()  # span
    ProvisionNum = auto()
    StatuteNum = auto()
    StatuteNamed = auto()
    StatutoryProvision = auto()  # span
    StatutoryLink = auto()  # span
    CaseName = auto()
    DocketNum = auto()
    ReporterNum = auto()
    DecisionCitation = auto()  # span
    GenericNum = auto()
    Candidate = auto()  # span
    CandidateProvision = auto()  # span

    @property
    def node(self) -> dict[str, str]:
        return {"ENT_TYPE": self.name}

    @property
    def opt(self) -> dict[str, str]:
        return self.node | {"OP": "?"}

    @property
    def snakecase(self) -> str:
        """Used as the variable for spacy's custom attributes e.g. `doc._.provision_nums`"""
        return "_".join(uncamel(self.name)).lower() + "s"

    def extract_entities(self, doc: Doc) -> list[Span]:
        """Generate doc's entities matching a label."""
        return [ent for ent in doc.ents if ent.label_ == self.name]

    def extract_spans(self, doc: Doc) -> list[Span]:
        """Generate doc's span ruler patterns matching a label."""
        return [span for span in filter_spans(doc.spans[self.name])]


SPAN_RULER_LABELS = [
    Label.Actor,
    Label.Concept,
    Label.Paper,
    Label.Doctrine,
    Label.DecisionCitation,
    Label.StatutoryProvision,
    Label.StatutoryLink,
    Label.Candidate,
    Label.CandidateProvision,
]
"""When included in this list, consider the `Rule` associated with the `Label` a _SpanRuler_ pattern."""


excludables = [Label.DATE, Label.ORDINAL] + SPAN_RULER_LABELS
"""No need to include these in the custom attributes of each Doc."""

ENTITY_RULER_LABELS = [member for member in Label if member not in excludables]
"""When included in this list, consider the `Rule` associated with the `Label` an _EntityRuler_ pattern."""


class Rule(BaseModel):
    """`patterns` associated with a single `label` (optionally
    with an `id` as well that serve as the `ent_id_` in `spacy.tokens.Doc.ents`).
    It can also be used for pattern objects for `spacy.tokens.Doc.spans`. See generally:

    1. https://spacy.io/usage/rule-based-matching#entityruler-files
    2. https://spacy.io/usage/rule-based-matching#spanruler-files

    A `Rule` enables the creation of such pattern objects containing the same `Label` and custom `id`, if provided. Sample rule:

    ```py
    sample = Rule(
        id="ministry-labor",
        label=Label.GovtDivision,
        patterns=[
            [
                {"LOWER": "the", "OP": "?"},
                {"LOWER": "ministry"},
                {"LOWER": "of"},
                {"LOWER": "labor"},
            ]
        ],
    )
    ```

    Use with: <Rule instance>.create_file() or <Rule instance>.model_dump() to
    simply get a list of patterns without file creation. See common usage in
    `set_entity_ruler()`
    """

    id: str | None = None
    label: Label
    patterns: list[list[dict[str, Any]]]

    def __str__(self) -> str:
        return f"<Rule {self.ent_id})>"

    def __repr__(self) -> str:
        return f"<Rule {self.ent_id}: {len(self.patterns)} patterns>"

    @property
    def ent_id(self):
        return self.id or "-".join(uncamel(self.label.name)).lower()

    @model_serializer
    def ser_model(self) -> list[dict[str, Any]]:
        """Following the pydantic convention for .model_dump(); instead of a traditional
        `dict` return, the function results in a serialized list of patterns for consumption
        by either `create_file()` or the _entity_ruler_ spacy pipeline."""
        return [
            {"id": self.ent_id, "label": self.label.name, "pattern": pattern}
            for pattern in self.patterns
        ]

    def create_file(self, file: Path | None = None):
        """Will update the file, if it exists; will create a file, if it doesn't exist."""
        if not file:
            file = Path(__file__).parent.joinpath(f"{self.ent_id}.jsonl")
        fp = io.BytesIO()
        with jsonlines.Writer(fp) as writer:
            for ser_pattern in self.model_dump():
                writer.write(ser_pattern)
        file.unlink(missing_ok=True)
        file.write_bytes(fp.getbuffer())
        return file

    @classmethod
    def extract_from_files(
        cls,
        folder: Path,
        glob_pattern: str = "*.py",
        excluded: tuple = ("__init__", "_pattern", "tester"),
        ents: bool = True,
    ) -> Iterator["Rule"]:
        """If the `Rule` instances found in the folder containg the glob pattern
        have a `.label` value that is True, then the rule should be mapped to the entity ruler;
        else, the rule should be mapped to the span ruler.

        The method of dynamically extracting from files is adopted from:
        https://stackoverflow.com/a/50181975/9081369

        Args:
            ents (bool, optional): See `Label` object whose value is `True` or `False`. Defaults to `True`.

        Yields:
            Iterator[Rule]: Rules that match the `ents` configuration.
        """
        for file in folder.glob(glob_pattern):
            if file.stem not in excluded:
                modules = {}
                spec = importlib.util.spec_from_file_location(file.stem, file)  # type: ignore
                modules[file.stem] = importlib.util.module_from_spec(spec)  # type: ignore
                spec.loader.exec_module(modules[file.stem])  # type: ignore
                for var_name, rule_candidate in modules[file.stem].__dict__.items():
                    if not var_name.startswith("__"):
                        if isinstance(rule_candidate, Rule):
                            if not rule_candidate.id:
                                rule_candidate.id = slugify(var_name)
                            if ents:
                                if rule_candidate.label not in SPAN_RULER_LABELS:
                                    yield rule_candidate
                            else:
                                if rule_candidate.label in SPAN_RULER_LABELS:
                                    yield rule_candidate


def _re(v: str, anchored: bool = True) -> dict[str, Any]:
    """Helper function to add an anchored, i.e. `^`<insert value `v` here>`$`
    regex pattern, following the convention `{"TEXT": {"REGEX": f"^{v}$"}}`
    spacy convention.
    """
    if anchored:
        v = f"^{v}$"
    return {"TEXT": {"REGEX": v}}


def _orth_in(v_list: list[str]):
    """Helper function to add a regex pattern where the options contained
    in the `v_list` are limited, following the convention `{"ORTH": {"IN": v_list}}`
    spacy convention.
    """
    return {"ORTH": {"IN": v_list}}


def _multi(pattern: list[dict[str, Any]], count: int) -> list[list[dict[str, Any]]]:
    """Create cumulative `pattern`s based on the `count` indicated.
    e.g. `repeat_pattern([1,2])` results in:

    [
        [1, 2],
        [1, 2, 1, 2],
        [1, 2, 1, 2, 1, 2],
        [1, 2, 1, 2, 1, 2, 1, 2],
        [1, 2, 1, 2, 1, 2, 1, 2, 1, 2]
    ]
    ```
    """
    res: list[list[dict[str, Any]]] = []
    res.append(pattern)
    for i in range(count):
        added = pattern + res[-1]
        res.append(added)
    return res


# REUSABLES


OF = {"LOWER": "of"}
"""Required `the`: `{"LOWER": "of"}`"""

TH = {"LOWER": "the", "OP": "?"}
"""Optional `the`: `{"LOWER": "the", "OP": "?"}`"""

COURT = {"ORTH": "Court"}
"""Required title cased `Court`"""

CODE = {"ORTH": "Code"}
"""Required title cased `Code`"""

THE = {"LOWER": "the"}
"""Required `the`: `{"LOWER": "the"}`"""

VS = _orth_in(["v.", "vs."])
"""Common formula for consistency: `{"ORTH": {"IN": ["v.", "vs."]}}`"""

PH = _orth_in(["Phil", "Phil.", "Phils", "Phils.", "Philippines"])

CM = {"ORTH": ","}

OF_THE_PH_ = [OF, THE, PH]

CONNECTOR = _orth_in(["of", "the", ",", "and", "&"]) | {"OP": "*"}
"""Setting this optional token ("of", "the", ",", "and", "&") allows for
ProvisionNums/serials to be merged later"""

### PATTERN FUNCS


def lower_words(words: str):
    """Will separate each word separated by spaces into a
    `{"LOWER": <word.strip().lower()>}` spacy pattern."""
    return [{"LOWER": word.strip().lower()} for word in words.split()]


def titled_words(words: str):
    """Will separate each word separated by spaces into a
    `{"LOWER": <word.lower()>, "IS_TITLE": True}` spacy pattern."""
    return [w | {"IS_TITLE": True} for w in lower_words(words)]


def name_code(fragments: str):
    """Use of `lower_words()` for Code names"""
    return [TH, {"LOWER": "code"}, OF] + lower_words(fragments)


def name_court(fragments: str):
    """Create title-based Court names"""
    return [TH, COURT, OF] + titled_words(fragments)


def name_statute(
    fragments: str | list[dict[str, Any]],
    options: list[str] = ["law", "act"],
) -> list[dict[str, Any]]:
    """If a string is passed as fragments, uses `lower_words()`
    in between options for named laws; if what is passed is a list of
    these same spacy patterns, these will be added as options."""
    bits: list[dict[str, Any]] = [TH]
    if isinstance(fragments, str):
        bits.extend(lower_words(fragments))
    elif isinstance(fragments, list):
        bits.extend(fragments)
    bits.append({"LOWER": {"IN": options}, "IS_TITLE": True})
    return bits


## DIGITS


p = inflect.engine()

roman_upper = [roman.toRoman(i) for i in range(1, 100)]
a_z_lower = [i for i in string.ascii_lowercase]
pairs = itertools.combinations_with_replacement(a_z_lower, r=2)
aa_zz_lower = [f"{a}{b}" for a, b in pairs]


class DigitLists(Enum):
    HundredDigit = [str(i) for i in range(0, 100)]
    WordHundredDigit = [p.number_to_words(num=i) for i in range(1, 100)]  # type: ignore
    RomanHundredLower = [i.lower() for i in roman_upper]
    RomanHundredUpper = [roman.toRoman(i) for i in range(1, 100)]
    AtoZSingleLower = a_z_lower
    AtoZSingleUpper = [i.upper() for i in a_z_lower]
    AtoZDoubleLower = aa_zz_lower
    AtoZDoubleUpper = [i.upper() for i in aa_zz_lower]

    @classmethod
    def generate_options(cls) -> list[str]:
        options: list[str] = []
        for member in cls:
            options.extend(member.value)  # type: ignore
        return options


HAS_DIGIT = _re(v=".*\\d.*")
"""Any token containing a digit should be used in tandem with an attribute ruler."""

IS_COVERED = _re(v=".*\\(\\w{1,2}\\).*")
"""Any token containing a digit should be used in tandem with an attribute ruler."""

SPECIFIC = _re("(" + "|".join(DigitLists.generate_options()) + ")")
"""Any token matching the options created by DigitLists"""

IS_ROMAN = _re(v="[IXV]+[-\\.][A-Z]{1,2}")
"""Handle combinations like I-A"""

PROV_DIGITS = [SPECIFIC, HAS_DIGIT, IS_COVERED, IS_ROMAN]



import spacy
from spacy.lang.char_classes import (
    ALPHA,
    ALPHA_LOWER,
    ALPHA_UPPER,
    CONCAT_QUOTES,
    LIST_ELLIPSES,
    LIST_ICONS,
)

from lextok.rules.abbreviations import Abbv, Prov
from lextok.rules.entity_rules_citeable import DocketNum, StatuteNum

# Remove hyphen '-' as infix, see https://spacy.io/usage/linguistic-features#native-tokenizer-additions
INFIXES_OVERRIDE = (
    LIST_ELLIPSES
    + LIST_ICONS
    + [
        r"(?<=[0-9])[+\\-\\*^](?=[0-9-])",
        r"(?<=[{al}{q}])\\.(?=[{au}{q}])".format(
            al=ALPHA_LOWER, au=ALPHA_UPPER, q=CONCAT_QUOTES
        ),
        r"(?<=[{a}]),(?=[{a}])".format(a=ALPHA),
        # ✅ Commented out regex that splits on hyphens between letters:
        # r"(?<=[{a}])(?:{h})(?=[{a}])".format(a=ALPHA, h=HYPHENS),
        r"(?<=[{a}0-9])[:<>=/](?=[{a}])".format(a=ALPHA),
    ]
)


def custom_prefix_list(nlp: spacy.language.Language):
    """Only use prefix `(` and `[` if _not_ followed by a single word `\\w+`
    with a closing `)` or `]`

    Note that modifications to a prefix should be done after the prefix removed, e.g.
    if the prefix is `(`, modify the `(`<add here> when appending a new rule.
    This is because of `compile_suffix_regex` which appends a `^` at the start of every prefix.

    example | status of `(`
    --:|:--
    `(`Juan de la Cruz v. Example) | is prefix
    Juan `(`de) la Cruz v. Example | is _not_ prefix
    """
    pre = list(nlp.Defaults.prefixes)  # type: ignore
    pre.remove("\\(")
    pre.append("\\((?![\\w\\.]+\\))")

    pre.remove("\\[")
    pre.append("\\[(?![\\w\\.]+\\])")
    return pre


def custom_suffix_list(nlp: spacy.language.Language):
    sfx = list(nlp.Defaults.suffixes)  # type: ignore
    sfx.remove("\\)")
    sfx.remove("\\]")
    return sfx


def create_special_rules():
    def make_special(texts: list[str]):
        """Add a period after every text item in `texts`, to consider each a single token.
        These patterns can be used as a special rule in creating a custom tokenizer."""
        return {f"{t}.": [{"ORTH": f"{t}."}] for t in texts if not t.endswith(".")}

    x = make_special("vs v s et al etc Ll Pp PP P.P R.P H.B S.B a.k.a".split())

    a = {
        k: v
        for member in DocketNum
        if member.value.initials
        for k, v in member.value.initials.as_token.items()
    }

    b = {
        k: v
        for member in StatuteNum
        if member.value.initials
        for k, v in member.value.initials.as_token.items()
    }

    c = {
        f"{bit}.": [{"ORTH": f"{bit}."}]
        for style in (None, "lower", "upper")
        for bit in Abbv.set_abbvs(cased=style)
    }

    d = {
        f"{bit}.": [{"ORTH": f"{bit}."}]
        for style in (None, "lower", "upper")
        for bit in Prov.set_abbvs(cased=style)
    }

    return a | b | c | d | x



from typing import Any

from lextok.rules._pattern import (
    CM,
    COURT,
    OF,
    OF_THE_PH_,
    TH,
    Label,
    Rule,
    _orth_in,
    _re,
    lower_words,
    titled_words,
)
from lextok.rules.abbreviations import CaseName, abbv_months, org_options
from lextok.rules.entity_rules_citeable import num_sym, opt_nums, provs

casename = Rule(label=Label.CaseName, patterns=CaseName.permute_patterns())

date_as_entity = Rule(
    label=Label.DATE,
    patterns=[
        [
            CM | {"OP": "?"},
            {"LOWER": {"IN": ["s.", "series"]}},
            OF | {"OP": "?"},
            _re("\\d{4}"),
        ]
    ]
    + [
        [
            CM | {"OP": "?"},
            {"ORTH": month},
            _orth_in([f"{str(i)}" for i in range(1, 31)]),
            CM | {"OP": "?"},
            _re("\\d{4}"),
        ]
        for month_data in abbv_months
        for month in month_data.value.options
    ],
)
prefix_tits = _orth_in(["Atty.", "Hon.", "Engr.", "Dr.", "Dra."])
prefix_titled_person = Rule(
    label=Label.PERSON,
    patterns=[
        [prefix_tits, {"IS_TITLE": True, "OP": "+"}],
        [prefix_tits, {"ENT_TYPE": Label.PERSON.name}],
    ],
)


juridical_org = Rule(
    label=Label.ORG,
    patterns=[
        [
            {"IS_TITLE": True, "OP": "+"},
            {"ORTH": ",", "IS_PUNCT": True, "OP": "?"},
            {
                "LOWER": {"IN": list(set(o.lower() for o in org_options))},
                "IS_TITLE": True,
            },
        ],
        [{"ORTH": "Estate"}, OF, {"IS_TITLE": True, "OP": "+"}],
        [{"ORTH": "Estate"}, OF, {"ENT_TYPE": Label.PERSON.name}],
    ],
)

generic_laws = Rule(
    label=Label.StatuteNamed,
    patterns=[
        [
            TH,
            _orth_in(["1987", "1973", "1935"]) | {"OP": "?"},
            _orth_in(["CONSTITUTION", "Constitution", "CONST", "Const."]),
        ],
        [{"ORTH": "Rules"}, OF, COURT],
        titled_words("old code"),
        titled_words("philippine civil code"),
        titled_words("civil code of 1950"),
        titled_words("civil code of 1889"),
        titled_words("civil code"),
        titled_words("civil code") + OF_THE_PH_,
        lower_words("revised administrative code of 1987"),
        lower_words("revised administrative code of 1917"),
        lower_words("revised administrative code"),
        lower_words("administrative code of 1987"),
        lower_words("administrative code of 1917"),
        lower_words("administrative code of 1916"),
        lower_words("administrative code"),
        lower_words("admin. code"),
        lower_words("admin code"),
        lower_words("adm. code"),
        lower_words("adm code"),
    ],
)


up = {"IS_UPPER": True, "OP": "+"}
tit = {"IS_TITLE": True, "OP": "+"}
opt_tit = {"IS_TITLE": True, "OP": "*"}

opt_cover = {"TEXT": {"REGEX": "\\(.*\\)"}, "OP": "?"}

doc_variants: list[list[dict[str, Any]]] = [
    [up, tit, opt_cover, Label.GenericNum.node, opt_nums, Label.DATE.node],
    [up, opt_cover, Label.GenericNum.node, opt_nums, Label.DATE.node],
    [tit, opt_cover, Label.GenericNum.node, opt_nums, Label.DATE.node],
    [up, tit, opt_cover, Label.GenericNum.node],
    [up, opt_cover, Label.GenericNum.node],
    [tit, opt_cover, Label.GenericNum.node],
    [up, tit, Label.GenericNum.node],
    [up, Label.GenericNum.node],
    [tit, Label.GenericNum.node],
    [Label.ORG.node, Label.GenericNum.node],
]
generic_documents = Rule(label=Label.Candidate, patterns=doc_variants)
"""Connect a generic serial number with a date, e.g. `Doc No. 414, Jan. 15, 2000`"""


irr_variants: list[list[dict[str, Any]]] = [
    [
        {"LOWER": "rules", "IS_TITLE": True},
        {"LOWER": "and", "IS_LOWER": True},
        {"LOWER": "regulations", "IS_TITLE": True},
        {"LOWER": "implementing", "IS_TITLE": True},
        {"IS_TITLE": True, "OP": "*"},
    ],
    [
        {"LOWER": "implementing", "IS_TITLE": True},
        {"LOWER": "rules", "IS_TITLE": True},
        {"LOWER": "and", "IS_LOWER": True},
        {"LOWER": "regulations", "IS_TITLE": True},
        {"ORTH": {"IN": ["of", "the"]}, "OP": "*"},
        {"IS_TITLE": True, "OP": "*"},
    ],
    [
        {"LOWER": {"IN": ["old", "new", "pre"]}, "IS_LOWER": True, "OP": "?"},
        {"ORTH": "-", "OP": "?"},
        {"SHAPE": "dddd", "OP": "?"},  # 1997, 1964, 2004
        {"IS_UPPER": True, "OP": "*"},  # NLRC, COMELEC, DARAB, Revised
        {"IS_TITLE": True, "OP": "*"},  # Revised
        {"LOWER": {"IN": ["rule", "rules"]}, "IS_TITLE": True},
        {"LOWER": {"IN": ["on", "of"]}},
        {"IS_TITLE": True, "OP": "*"},
    ],
    [
        {"SHAPE": "dddd", "OP": "?"},  # 1997, 1964, 2004
        {"IS_UPPER": True, "OP": "*"},  # NLRC
        {"IS_TITLE": True, "OP": "*"},  # Revised
        {"LOWER": "revised", "OP": "?"},
        {
            "LOWER": {"IN": ["omnibus", "uniform", "implementing", "rules", "interim"]},
            "IS_TITLE": True,
            "OP": "+",
        },
    ],
]
opt_connect: dict[str, Any] = {"ORTH": {"IN": ["of", "the", ","]}, "OP": "*"}

generic_irrs = Rule(label=Label.Candidate, patterns=irr_variants)

prov_irr = Rule(
    label=Label.CandidateProvision,
    patterns=[[provs, opt_connect] + i for i in irr_variants]
    + [[provs, opt_connect] + d for d in doc_variants],
)



from typing import Any, NamedTuple

from lextok.rules._pattern import _orth_in, _re


class Duo(NamedTuple):
    """Given two letters, create possible patterns using uppercase text."""

    a: str
    b: str

    @property
    def x(self):
        return self.a.upper()

    @property
    def y(self):
        return self.b.upper()

    @property
    def as_token(self) -> dict[str, list[dict[str, str]]]:
        """Used to create special rules for custom tokenizer."""
        return {f"{self.x}.{self.y}.": [{"ORTH": f"{self.x}.{self.y}."}]}

    @property
    def v1(self) -> list[dict[str, str]]:
        # R . A .
        return [{"ORTH": self.x}, {"ORTH": "."}, {"ORTH": self.y}, {"ORTH": "."}]

    @property
    def v2(self) -> list[dict[str, str]]:
        return [{"ORTH": f"{self.x}."}, {"ORTH": f"{self.y}."}]  # R. A.

    @property
    def v3(self) -> list[dict[str, str]]:
        return [{"ORTH": f"{self.x}.{self.y}."}]  # R.A.

    @property
    def v4(self) -> list[dict[str, str]]:
        return [{"ORTH": f"{self.x}{self.y}"}]  # RA

    @property
    def patterns(self) -> list[list[dict[str, str]]]:
        return [self.v1, self.v2, self.v3, self.v4]

    def add_to_each_pattern(self, terminators: list[dict[str, Any]]):
        for p in self.patterns:
            yield p + terminators


class Style(NamedTuple):
    """A `Style` refers to the way a specific two-`let`ter acronym can be used to create
    various letter patterns along with the word variants described in `v`.
    If no `let` exists, will only create word patterns using `v`."""

    v: list[str] = []
    let: str | None = None
    pre: list[str] = ["No", "Nos", "No.", "Nos."]
    r: str = "[\\w-]+"

    @property
    def title_pre(self) -> list[dict[str, Any]]:
        return [_orth_in(self.pre), _re(self.r)]

    @property
    def upper_pre(self) -> list[dict[str, Any]]:
        return [_orth_in([i.upper() for i in self.pre]), _re(self.r)]

    @property
    def token_parts(self):
        """The first pass is for indiscriminate words e.g. `bar matter`; the second, for
        dealing with periods, e.g. `adm. matter`. The first will generate the following as
        token parts ('bar','matter'); the second: ('adm.','matter'), ('adm','.','matter')
        """
        objs = set()
        for words in self.v:
            partial = []
            for word in words.split():
                partial.append(word)
            objs.add(tuple(partial))
        for words in self.v:
            partial = []
            for word in words.split():
                if word.endswith("."):
                    cleaned = word.removesuffix(".")
                    partial.append(cleaned)
                    partial.append(".")
                else:
                    partial.append(word)
            objs.add(tuple(partial))
        return objs

    @property
    def _title_num(self) -> list[list[dict[str, Any]]]:
        return [
            [{"ORTH": sub.title()} for sub in subtokens] + self.title_pre
            for subtokens in self.token_parts
        ]

    @property
    def _title_no_num(self) -> list[list[dict[str, Any]]]:
        return [
            [{"ORTH": sub.title()} for sub in subtokens] + [_re(self.r)]
            for subtokens in self.token_parts
        ]

    @property
    def _upper_num(self) -> list[list[dict[str, Any]]]:
        return [
            [{"ORTH": sub.upper()} for sub in subtokens] + self.upper_pre
            for subtokens in self.token_parts
        ]

    @property
    def _upper_no_num(self) -> list[list[dict[str, Any]]]:
        return [
            [{"ORTH": sub.upper()} for sub in subtokens] + [_re(self.r)]
            for subtokens in self.token_parts
        ]

    @property
    def initials(self) -> Duo | None:
        if not self.let:
            return None
        if len(self.let) != 2:
            return None
        return Duo(a=self.let[0], b=self.let[1])

    @property
    def word_patterns(self) -> list[list[dict[str, Any]]]:
        patterns = []
        patterns.extend(self._title_num)
        patterns.extend(self._title_no_num)
        patterns.extend(self._upper_num)
        patterns.extend(self._upper_no_num)
        return patterns

    @property
    def letter_patterns(self) -> list[list[dict[str, Any]]]:
        items: list[list[dict[str, Any]]] = []
        if not self.initials:
            return items
        for target_nodes in (self.upper_pre, self.title_pre, [_re(self.r)]):
            for b in self.initials.add_to_each_pattern(target_nodes):
                items.append(b)
        return items

    @property
    def patterns(self) -> list[list[dict[str, Any]]]:
        words = self.word_patterns
        letters = self.letter_patterns
        return words + letters if letters else words



from .main import (
    ENTITY_RULES,
    SPAN_RULES,
    EntitySpanDetectorComponent,
    create_custom_entities,
    create_custom_spans,
    lextok,
    merge_provisions,
    set_attribute_ruler,
    set_tokenizer,
)
from .pretest import pretest_entities
from .rules import (
    CM,
    CODE,
    CONNECTOR,
    COURT,
    OF,
    OF_THE_PH_,
    TH,
    THE,
    Abbv,
    Attr,
    CourtName,
    DigitLists,
    DocketNum,
    Duo,
    Label,
    Prov,
    ProvisionNum,
    ReporterNum,
    Rule,
    StatuteNum,
    Style,
    WordAttributes,
    _orth_in,
    _re,
    lower_words,
    name_code,
    name_court,
    name_statute,
    titled_words,
    uncamel,
)



import itertools
from collections.abc import Iterator
from enum import Enum
from typing import Any, NamedTuple

from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

from lextok.rules._pattern import PROV_DIGITS, VS, Label, _orth_in, _re


def check_titlecased_word(v: str) -> str:
    assert all(bit.istitle for bit in v.split("-")), f"{v} is not titlecased."
    return v


TitledString = Annotated[str, AfterValidator(check_titlecased_word)]


class Def(NamedTuple):
    """A (possible) definition of a commonly used abbreviation;
    each value must be a `TitledString` so that it can be
    adjusted for lowercase / uppercase variants."""

    title: TitledString
    abbv: TitledString | None = None

    @staticmethod
    def get_cased_value(v: str, cased: str | None = None) -> str:
        if cased:
            if cased == "lower":
                return v.lower()
            elif cased == "upper":
                return v.upper()
        return v

    @property
    def dotted_abbv(self) -> list[str]:
        bits = []
        if self.abbv:
            for style in (None, "lower", "upper"):
                bits.append(self.get_cased_value(self.abbv, cased=style))
                bits.append(self.get_cased_value(self.abbv + ".", cased=style))
        return bits

    @property
    def options(self) -> list[str]:
        bits = []
        for style in (None, "lower", "upper"):
            bits.append(self.get_cased_value(self.title, cased=style))
        return bits + self.dotted_abbv


class Abbv(Enum):
    """Some common abbreviations used in Philippine legal text."""

    Adm = Def(title="Administrative", abbv="Adm")
    Admin = Def(title="Administrative", abbv="Admin")
    Pres = Def(title="Presidential", abbv="Pres")
    Dec = Def(title="Decree", abbv="Dec")
    Executive = Def(title="Executive", abbv="Exec")
    Blg = Def(title="Bilang", abbv="Blg")
    Number = Def(title="Number", abbv="No")
    Numbers = Def(title="Numbers", abbv="Nos")
    Const = Def(title="Constitution", abbv="Const")
    Company = Def(title="Company", abbv="Co")
    Corporation = Def(title="Corporation", abbv="Corp")
    Incorporated = Def(title="Incorporated", abbv="Inc")
    Phil1 = Def(title="Philippines", abbv="Phil")
    Phil2 = Def(title="Philippines", abbv="Phils")
    Limited = Def(title="Limited", abbv="Ltd")
    Association = Def(title="Association", abbv="Assoc")
    Assistant = Def(title="Assistant", abbv="Ass")
    Department = Def(title="Department", abbv="Dept")
    Nat1 = Def(title="National", abbv="Nat")
    Nat2 = Def(title="National", abbv="Natl")
    St = Def(title="Street", abbv="St")
    Road = Def(title="Road", abbv="Rd")
    Ave = Def(title="Avenue", abbv="Ave")
    Blk = Def(title="Block", abbv="Blk")
    Brgy = Def(title="Barangay", abbv="Brgy")
    Building = Def(title="Building", abbv="Bldg")
    Purok = Def(title="Purok", abbv="Prk")
    Subdivision = Def(title="Subdivision", abbv="Subd")
    Highway = Def(title="Highway", abbv="Hwy")
    Municipality = Def(title="Municipality", abbv="Mun")
    City = Def(title="City", abbv="Cty")
    Province = Def(title="Province", abbv="Prov")
    Governor = Def(title="Governor", abbv="Gov")
    Congressman = Def(title="Congressman", abbv="Cong")
    General = Def(title="General", abbv="Gen")
    Lieutenant = Def(title="Lieutenant", abbv="Lt")
    Sct = Def(title="Scout", abbv="Sct")
    Sta = Def(title="Santa", abbv="Sta")
    Sto = Def(title="Santo", abbv="Sto")
    Vda = Def(title="Viuda", abbv="Vda")
    Jr = Def(title="Junior", abbv="Jr")
    Sr = Def(title="Senior", abbv="Sr")
    Fr = Def(title="Father", abbv="Fr")
    Bro = Def(title="Brother", abbv="Bro")
    Dr = Def(title="Doctor", abbv="Dr")
    Dra = Def(title="Doctora", abbv="Dra")
    Maria = Def(title="Maria", abbv="Ma")
    Hon = Def(title="Honorable", abbv="Hon")
    Atty = Def(title="Attorney", abbv="Atty")
    Engr = Def(title="Engineer", abbv="Engr")
    Justice = Def(title="Justice", abbv="J")
    January = Def(title="January", abbv="Jan")
    February = Def(title="February", abbv="Feb")
    March = Def(title="March", abbv="Mar")
    April = Def(title="April", abbv="Apr")
    May = Def(title="May")
    June = Def(title="June", abbv="Jun")
    July = Def(title="July", abbv="Jul")
    August = Def(title="August", abbv="Aug")
    Sept1 = Def(title="September", abbv="Sept")
    Sept2 = Def(title="September", abbv="Sep")
    October = Def(title="October", abbv="Oct")
    November = Def(title="November", abbv="Nov")
    December = Def(title="December", abbv="Dec")

    @classmethod
    def set_abbvs(cls, cased: str | None = None) -> Iterator[str]:
        for member in cls:
            if v := member.value.abbv:
                yield Def.get_cased_value(v, cased)

    @classmethod
    def set_fulls(cls, cased: str | None = None) -> Iterator[str]:
        for member in cls:
            yield Def.get_cased_value(member.value.title, cased)


class Prov(Enum):
    """A ProvisionNum of a statute may be abbreviated and the same may have different variations:
    e.g. titlecase, lowercase, and uppercase."""

    Title = Def(title="Title", abbv="Tit")
    SubT0 = Def(title="Subtitle")
    SubT1 = Def(title="SubTitle")
    SubT2 = Def(title="Sub-Title")
    Book = Def(title="Book", abbv="Bk")
    Chapter = Def(title="Chapter", abbv="Ch")
    SubChap0 = Def(title="Subchapter")
    SubChap1 = Def(title="SubChapter")
    SubChap2 = Def(title="Sub-Chapter", abbv="Sub-Chap")
    Article = Def(title="Article", abbv="Art")
    SubArt0 = Def(title="Subarticle")
    SubArt1 = Def(title="SubArticle")
    SubArt2 = Def(title="Sub-Article", abbv="Sub-Art")
    Section = Def(title="Section", abbv="Sec")
    SubSec0 = Def(title="Subsection")
    SubSec1 = Def(title="SubSection")
    SubSec2 = Def(title="Sub-Section", abbv="Sub-Sec")
    Par = Def(title="Paragraph", abbv="Par")
    SubPar0 = Def(title="Subparagraph")
    SubPar1 = Def(title="SubParagraph")
    SubPar2 = Def(title="Sub-Paragraph", abbv="Sub-Par")
    ProvRule = Def(title="Rule")
    Canon = Def(title="Canon")

    @classmethod
    def make_attr_rules(cls) -> Iterator[dict[str, Any]]:
        """Each member can contain explicit rules so that "digit patterns"
        are paired with the "adjective patterns":

        first node (adjective) | second node (digit)
        :-- | --:
        The member option for Sec., Section, etc. | a "digit" pattern e.g. 1, 1(a)

        The list of patterns can then be applied as part of an attribute ruler
        https://spacy.io/usage/linguistic-features#mappings-exceptions so that
        the token in the second node, i.e. the digit, can be set with the attributes defined in
        `attributes_to_set`.
        """
        for member in cls:
            p = [[{"ORTH": o}, p] for p in PROV_DIGITS for o in member.value.options]
            yield {"index": 0, "patterns": p, "attrs": {"POS": "NOUN"}}
            yield {"index": 1, "patterns": p, "attrs": {"POS": "NUM", "LIKE_NUM": True}}

    @classmethod
    def set_abbvs(cls, cased: str | None = None):
        for member in cls:
            if v := member.value.abbv:
                yield Def.get_cased_value(v, cased)

    @classmethod
    def set_fulls(cls, cased: str | None = None):
        for member in cls:
            yield Def.get_cased_value(member.value.title, cased)


abbv_months = (
    Abbv.January,
    Abbv.February,
    Abbv.March,
    Abbv.April,
    Abbv.June,
    Abbv.July,
    Abbv.August,
    Abbv.Sept1,
    Abbv.Sept2,
    Abbv.October,
    Abbv.November,
    Abbv.December,
)

org_options = set(
    op
    for s in (Abbv.Company, Abbv.Corporation, Abbv.Limited, Abbv.Incorporated)
    for op in s.value.options
)
ph_options = set(op for s in (Abbv.Phil1, Abbv.Phil2) for op in s.value.options)
extras = " ".join(org_options)
extras += " " + " ".join(ph_options)
extras += " . , et al. et al the Jr Jr. Sr Sr. III IV Partnership Dev't"
misc = _orth_in(extras.split()) | {"OP": "*"}
opt_acronym = _re("\\([A-Z]+\\)") | {"OP": "?"}


class CaseName(Enum):
    ent = {
        "ENT_TYPE": {
            "IN": [
                Label.ORG.name,
                Label.PERSON.name,
                Label.LAW.name,
            ]
        },
        "OP": "+",
    }
    pos = {"POS": {"IN": ["PROPN", "ADP", "DET"]}, "OP": "+"}
    upper = {"IS_UPPER": True, "OP": "{1,6}"}
    title = {"IS_TITLE": True, "OP": "{1,6}"}

    @staticmethod
    def _vs(first: dict[str, Any], second: dict[str, Any]) -> list[dict[str, Any]]:
        return [first, opt_acronym, misc] + [VS] + [second, opt_acronym, misc]

    @classmethod
    def permute_patterns(cls) -> list[list[dict[str, Any]]]:
        pairs = itertools.permutations((m.value for m in cls), 2)
        return [cls._vs(a, b) for a, b in pairs]



from spacy.language import Language
from spacy.tokens.doc import Doc


def pretest_entities(
    raw_nlp: Language,
    passages: list[str],
    test_ent_id: str = "XXX",
    patterns: list[list[dict]] | None = None,
) -> list[list[str]]:
    """All `passages` should result in a matching pattern
    from the list of possible `patterns`"""
    matches: list[list[str]] = []
    if not passages:
        return matches
    try:
        raw_nlp.remove_pipe("entity_ruler")
    except ValueError:
        pass

    if patterns:
        ruler = raw_nlp.add_pipe(
            "entity_ruler",
            config={"overwrite_ents": True},
            validate=True,
        )
        with raw_nlp.select_pipes(enable="tagger"):
            for pattern in patterns:
                ruler.add_patterns([{"id": test_ent_id, "label": "test", "pattern": pattern}])  # type: ignore

    for passage in passages:
        doc: Doc = raw_nlp(passage)  # type: ignore
        found = False
        passage_matches: list[str] = []
        for ent in doc.ents:
            if ent.ent_id_ == test_ent_id:
                found = True
                passage_matches.append(ent.text)
        if not found:
            raise Exception(f"Undetected {id}: {passage}")
        matches.append(passage_matches)

    return matches



from enum import Enum
from typing import Any, NamedTuple

from lextok.rules._pattern import CM, OF, THE, VS, Label, _orth_in, _re
from lextok.rules.abbreviations import Abbv


class Attr(NamedTuple):
    """Creates a repeatable pattern for spacy's `AttributeRuler` in the customized
    tokenizer: a pattern is a list of tokens where each token is represented by the
    index. The specified index in `indexes` means that the user desires to change
    the attribute of the token represented by the index with the attribute values
    set in `attrs`. See example in https://spacy.io/usage/linguistic-features#mappings-exceptions
    """

    pattern: list[dict[str, Any]]  # the list of tokens
    attrs: dict[str, Any]  # e.g. make index 0 of the `pattern`` a NUM
    indexes: list[int]  # default to the first token in the pattern
    examples: list[str] | None = None

    def create_rules(self):
        for index in self.indexes:
            yield {"patterns": [self.pattern], "attrs": self.attrs, "index": index}


class WordAttributes(Enum):
    num = Attr(
        pattern=[
            {"IS_TITLE": True},
            _orth_in(["No", "Nos", "No.", "Nos."]),
        ],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.LAW.name},
        indexes=[0, 1],
    )

    inc_a_with_comma = Attr(
        pattern=[
            {"IS_ALPHA": True},
            CM,
            _orth_in(
                Abbv.Company.value.options
                + Abbv.Corporation.value.options
                + Abbv.Limited.value.options
            ),
        ],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.ORG.name},
        indexes=[0, 1, 2],
    )

    inc_b_sans_comma = Attr(
        pattern=[
            {"IS_ALPHA": True},
            _orth_in(
                Abbv.Company.value.options
                + Abbv.Corporation.value.options
                + Abbv.Limited.value.options
            ),
        ],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.ORG.name},
        indexes=[0, 1],
    )

    etal = Attr(
        pattern=[
            {"IS_TITLE": True},
            CM,
            {"LOWER": {"IN": ["et", "et."]}},
            {"LOWER": {"IN": ["al", "al."]}},
        ],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.ORG.name},
        indexes=[0, 1, 2, 3],
    )

    person_prefix = Attr(
        pattern=[_orth_in(Abbv.Vda.value.options), {"LOWER": "de"}, {"IS_ALPHA": True}],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.PERSON.name},
        indexes=[0, 1, 2],
    )

    de_la = Attr(
        pattern=[{"LOWER": "de"}, {"LOWER": {"IN": ["la", "leon", "los"]}}],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.PERSON.name},
        indexes=[0, 1],
    )

    dela = Attr(
        pattern=[{"ORTH": {"IN": ["de", "De", "dela", "Dela", "Delos"]}}],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.PERSON.name},
        indexes=[0],
    )

    rp = Attr(
        pattern=[
            {
                "LOWER": {
                    "IN": (
                        "rep rep. republic people pp p.p. pp. govt govt. government"
                        " gov't".split()
                    )
                }
            },
            OF,
            THE,
            {"LOWER": {"IN": "phil phils phil. phils. philippines".split()}},
        ],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.ORG.name},
        indexes=[0, 1, 2, 3],
    )

    people_v = Attr(
        pattern=[{"LOWER": {"IN": ["pp", "pp.", "people", "republic", "rp"]}}, VS],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.ORG.name},
        indexes=[0],
    )

    v_court = Attr(
        pattern=[VS, {"ORTH": {"IN": ["CA", "CFI", "Sandiganbayan"]}}],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.ORG.name},
        indexes=[1],
    )

    dot_pre_v = Attr(
        pattern=[{"IS_ASCII": True}, {"ORTH": "."}, VS],
        attrs={"POS": "PROPN", "IS_PUNCT": False},
        indexes=[0, 1],
    )

    person_suffix_pre_v = Attr(
        pattern=[CM, {"ORTH": {"IN": ["Jr.", "Sr."]}}, VS],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.PERSON.name},
        indexes=[0, 1],
    )

    person_suffix_pre_v2 = Attr(
        pattern=[
            {"IS_TITLE": True},
            {"ORTH": {"IN": ["Jr.", "Sr.", "II", "III", "IV", "V", "VI"]}},
            VS,
        ],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.PERSON.name},
        indexes=[0, 1],
    )

    v = Attr(
        pattern=[VS],
        attrs={"POS": "CCONJ", "ENT_TYPE": Label.LAW.name},
        indexes=[0],
    )

    v_parties1 = Attr(
        pattern=[{"IS_TITLE": True}, VS, {"IS_TITLE": True}],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.LAW.name},
        indexes=[0, 2],
    )

    v_parties2 = Attr(
        pattern=[{"IS_UPPER": True}, VS, {"IS_UPPER": True}],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.LAW.name},
        indexes=[0, 2],
    )

    v_parties3 = Attr(
        pattern=[{"IS_UPPER": True}, VS, {"IS_TITLE": True}],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.LAW.name},
        indexes=[0, 2],
    )

    v_parties4 = Attr(
        pattern=[{"IS_TITLE": True}, VS, {"IS_UPPER": True}],
        attrs={"POS": "PROPN", "ENT_TYPE": Label.LAW.name},
        indexes=[0, 2],
    )

    year = Attr(
        pattern=[_re(v="\\((194\\d|195\\d|196\\d|197\\d|198\\d|199\\d|20\\d{2})\\)")],
        attrs={"POS": "NUM", "ENT_TYPE": Label.DATE.name, "DEP": "nummod"},
        indexes=[0],
    )  # 1940s to 2000s wrapped around parenthesis

    @classmethod
    def make_attr_rules(cls):
        for member in cls:
            for rules in member.value.create_rules():
                yield rules  # type: ignore



from collections.abc import Iterator
from pathlib import Path

import spacy  # type: ignore
from spacy.language import Language
from spacy.tokenizer import Tokenizer  # type: ignore
from spacy.tokens import Doc, Span
from spacy.util import filter_spans

from .rules import (
    ENTITY_RULER_LABELS,
    INFIXES_OVERRIDE,
    SPAN_RULER_LABELS,
    Label,
    Prov,
    Rule,
    WordAttributes,
    create_special_rules,
    custom_prefix_list,
    custom_suffix_list,
)

RULES_DIR = Path(__file__).parent.joinpath("rules")

ENTITY_RULES = list(Rule.extract_from_files(folder=RULES_DIR, ents=True))

SPAN_RULES = list(Rule.extract_from_files(folder=RULES_DIR, ents=False))


def set_tokenizer(nlp: Language) -> Tokenizer:
    infix_re = spacy.util.compile_infix_regex(INFIXES_OVERRIDE)  # type: ignore
    suffix_re = spacy.util.compile_suffix_regex(custom_suffix_list(nlp))
    prefix_re = spacy.util.compile_prefix_regex(custom_prefix_list(nlp))
    return Tokenizer(
        nlp.vocab,
        rules=create_special_rules(),
        prefix_search=prefix_re.search,
        suffix_search=suffix_re.search,
        infix_finditer=infix_re.finditer,
    )


def set_attribute_ruler(nlp: Language):
    ruler = nlp.get_pipe("attribute_ruler")
    for rule in WordAttributes.make_attr_rules():
        ruler.add(**rule)  # type: ignore
    for rule in Prov.make_attr_rules():
        ruler.add(**rule)  # type: ignore
    return nlp


def create_custom_entities(
    nlp: Language, rules: list[Rule], pipename: str = "entity_ruler"
):
    ents = nlp.add_pipe(
        factory_name="entity_ruler",
        name=pipename,
        config={"overwrite_ents": True, "validate": True},
        validate=True,
    )
    for rule in rules:
        ents.add_patterns(rule.model_dump())  # type: ignore
    return nlp


def create_custom_spans(nlp: Language, rules: list[Rule]):
    for rule in rules:
        container = f"span_ruler_{rule.label.name}"
        try:
            spans = nlp.get_pipe(container)
        except KeyError:
            spans = nlp.add_pipe(
                "span_ruler",
                name=container,
                config={
                    "spans_key": rule.label.name,
                    "validate": True,
                },  # spans_key for accessing doc.spans[spans_key]
                validate=True,
            )
        spans.add_patterns(rule.model_dump())  # type: ignore
    return nlp


@Language.component("provision_num_merger")
def merge_provisions(doc: Doc) -> Doc:
    """Consecutive `ProvisionNum` entities merged as a _single_ `ProvisionNum` token and entity."""
    pairs = [(e.start, e.end) for e in doc.ents if e.label_ == Label.ProvisionNum.name]
    pair = None
    new_ents = []
    for ent in doc.ents:
        if ent.label_ == Label.ProvisionNum.name:
            s = ent.start
            if pair and s == pair[1]:
                s = pair[0]
            pair = (s, ent.end)
        if pair and pair not in pairs:
            new_ents.append(
                Span(doc=doc, start=pair[0], end=pair[1], label=Label.ProvisionNum.name)
            )
    if new_ents:
        doc.ents = filter_spans(new_ents + list(doc.ents))
    return doc


@Language.factory("detector")
class EntitySpanDetectorComponent:
    """Dynamic creation of custom attributes (`doc._.<attribute>`)  based on `Label` objects."""

    def __init__(self, nlp, name):
        labels = ENTITY_RULER_LABELS + SPAN_RULER_LABELS
        for label in labels:
            Doc.set_extension(label.snakecase, default=[])

    def __call__(self, doc):
        for label in ENTITY_RULER_LABELS:
            doc._.set(label.snakecase, label.extract_entities(doc))

        for label in SPAN_RULER_LABELS:
            try:
                doc._.set(label.snakecase, label.extract_spans(doc))
            except KeyError:
                continue
        return doc


def lextok(
    model: str = "en_core_web_sm",
    finalize_entities: bool = False,
    entity_rules: list[Rule] = ENTITY_RULES,
    span_rules: list[Rule] = SPAN_RULES,
) -> Language:
    """Incorporation of:

    1. overrides for `set_tokenizer()` and `set_attribute_ruler()`
    2. extendible entity ruler / span ruler patterns, see the `Rule` object
    3. last pipe `detector` which creates custom attributes on each `Doc`
    based on `Label`s detected.

    Args:
        model (str, optional): Spacy language model. Defaults to "en_core_web_sm".
        finalize_entities (bool, optional): Whether to consider consecutive entities as a single token. Defaults to False.
        entity_rules (list[Rule], optional): List of generic Pydantic models with serialization. Defaults to `ENTITY_RULES`.
        span_rules (list[Rule], optional): List of generic Pydantic models with serialization. Defaults to `SPAN_RULES`.

    Returns:
        Language: A customized rule-based spacy model.
    """
    nlp = spacy.load(model)
    nlp.tokenizer = set_tokenizer(nlp)
    nlp = set_attribute_ruler(nlp)
    create_custom_entities(nlp, rules=entity_rules, pipename="entity_ruler")
    if finalize_entities:
        nlp.add_pipe("merge_entities", after="entity_ruler")
        nlp.add_pipe("provision_num_merger", after="merge_entities")
        create_custom_spans(nlp, rules=span_rules)
        nlp.add_pipe("detector", last=True)
    return nlp
